use crate::store::{
    nsobject::NsObject,
    structs::{NsObjPath, ObjPath},
};
use crate::utils::path;
use std::env;
use std::fs::{self, File, OpenOptions};
use std::io::{self, Write};
use std::path::PathBuf;
use std::sync::OnceLock;
use std::time::{SystemTime, UNIX_EPOCH};

const MAX_SIZE_TO_USE_HASH: u64 = 12 * 1024 * 1024;

pub struct ObjMetadata {
    size: u64,
    modified: Option<SystemTime>,
}

#[derive(PartialEq, Clone, Debug)]
#[repr(u8)]
pub enum ObjType {
    Obj,
    Blob,
    Tree,
}

impl From<ObjType> for u8 {
    fn from(variant: ObjType) -> Self {
        variant as u8
    }
}

impl TryFrom<u8> for ObjType {
    type Error = String;

    fn try_from(byte: u8) -> Result<Self, Self::Error> {
        match byte {
            0 => Ok(ObjType::Obj),
            1 => Ok(ObjType::Blob),
            2 => Ok(ObjType::Tree),
            _ => Err(format!("Invalid byte value: {}", byte)),
        }
    }
}

#[derive(PartialEq, Clone, Debug)]
pub enum ObjStatus {
    Undefined,
    Unchanged,
    Created,
    Modified,
    Moved,
    Copied,
    Deleted,
}

#[derive(Clone, Debug)]
pub struct Obj {
    obj_type: ObjType,
    status: OnceLock<ObjStatus>,
    /// path of the object from root
    obj_path: ObjPath,
}

impl Obj {
    pub fn from_local_path(path: &ObjPath) -> Self {
        let obj_path = path.clone();
        Obj {
            obj_type: {
                if path.abs().exists() {
                    if path.abs().is_dir() {
                        ObjType::Tree
                    } else {
                        ObjType::Blob
                    }
                } else {
                    ObjType::Obj
                }
            },
            status: OnceLock::new(),
            obj_path,
        }
    }

    pub fn get_status(&self) -> &ObjStatus {
        self.status.get_or_init(|| {
            let nsobj = self.get_nsobj();
            if !nsobj.exists() {
                return ObjStatus::Created;
            }
            if !self.obj_path.exists() {
                return ObjStatus::Deleted;
            }
            if *self != nsobj {
                return ObjStatus::Modified;
            }
            todo!("Read status");
        })
    }

    pub fn set_status(&mut self, status: ObjStatus) {
        self.status = OnceLock::from(status);
    }

    pub fn set_type(&mut self, obj_type: ObjType) {
        self.obj_type = obj_type;
    }

    pub fn is_new(&self) -> bool {
        self.get_status() == &ObjStatus::Created
    }

    pub fn get_obj_type(&self) -> &ObjType {
        &self.obj_type
    }

    pub fn get_obj_path(&self) -> &ObjPath {
        &self.obj_path
    }

    /// Return the path of the current object relatively to the path the
    /// command was executed from.
    ///
    /// * `repo_root`: the absolute repo's root path
    pub fn get_env_relative_path(&self, repo_root: &PathBuf) -> PathBuf {
        let binding = env::current_dir().unwrap();

        if let Ok(root_diff) = binding.strip_prefix(repo_root) {
            match self.obj_path.strip_prefix(&root_diff) {
                Ok(path) => {
                    if path == PathBuf::from("") {
                        PathBuf::from("./")
                    } else {
                        path.to_path_buf()
                    }
                }
                Err(_) => {
                    // if cannot strip prefix then we need to go up to the root
                    let mut res_path = PathBuf::new();
                    for _ in 0..path::get_level(&root_diff.to_path_buf()) {
                        res_path.push("..");
                    }
                    res_path.push(&*self.obj_path);
                    res_path
                }
            }
        } else {
            // if cannot strip prefix then we are at the repo's root
            self.obj_path.to_path_buf()
        }
    }

    pub fn cpy_path(&self) -> String {
        path::to_string(&self.obj_path)
    }

    fn get_nsobj(&self) -> NsObject {
        NsObject::from_local_path(&self.obj_path)
    }

    fn get_nsobj_path(&self) -> NsObjPath {
        NsObjPath::from(&self.obj_path)
    }

    pub fn save(&self) -> io::Result<()> {
        let nsobj_path = self.get_nsobj_path();
        let mut file = {
            if nsobj_path.exists() {
                todo!("Edition of object")
            } else {
                let mut nsobj_dir = nsobj_path.clone();
                nsobj_dir.pop();
                if !nsobj_dir.exists() {
                    std::fs::create_dir_all(nsobj_dir)?;
                }
                File::create(&nsobj_path)?
            }
        };

        // Write type
        file.write(&[self.obj_type.clone().into()])?;

        if self.obj_type == ObjType::Blob {
            if let Some(metadata) = self.get_metadata() {
                // Write size
                file.write(&metadata.size.to_le_bytes())?;

                // Write modified
                file.write(
                    &metadata
                        .modified
                        .expect(&format!(
                            "Expect 'modified' in metadata of {} to save obj",
                            self.cpy_path()
                        ))
                        .duration_since(UNIX_EPOCH)
                        .unwrap()
                        .as_secs()
                        .to_le_bytes(),
                )?;
            } else {
                todo!("Cannot load metadata")
            }
        }

        file.write_all(b"\n")?;
        // Write Path
        file.write_all(self.obj_path.to_str().unwrap().as_bytes())?;
        file.write_all(b"\n")?;

        file.flush()?;

        // Save itself inside its parent
        let mut parent_path = self.obj_path.clone();
        parent_path.pop();
        let parent_obj = Obj::from_local_path(&parent_path);
        parent_obj.add_child(&self)?;

        Ok(())
    }

    fn add_child(&self, child: &Obj) -> io::Result<()> {
        let mut file = OpenOptions::new()
            .write(true)
            .create(true)
            .append(true)
            .open(self.get_nsobj_path())?;

        let child_type = &[child.get_obj_type().clone().into()];
        let child_path = child.obj_path.to_str().unwrap().as_bytes();
        file.write(child_type)?;
        file.write(child_path)?;
        file.write(b"\n")?;
        file.flush()?;
        Ok(())
    }

    pub fn get_metadata(&self) -> Option<ObjMetadata> {
        let metadata = match fs::metadata(&*self.obj_path) {
            Ok(m) => m,
            Err(err) => {
                eprintln!(
                    "Failed to read metadata of {} ({})",
                    self.obj_path.display(),
                    err
                );
                return None;
            }
        };

        Some(ObjMetadata {
            size: metadata.len(),
            modified: metadata.modified().ok(),
        })
    }
}

impl PartialEq<NsObject> for Obj {
    fn eq(&self, other: &NsObject) -> bool {
        if self.obj_type != other.obj_type {
            eprintln!("Trying to compare different obj type");
            return false;
        }

        if self.obj_type == ObjType::Tree {
            panic!("fatal: cannot comparing folders");
        }

        let obj_metadata = self.get_metadata();
        let nsobj_metadata = other.get_metadata();

        // If not metadata exist considere it is not the same
        if obj_metadata.is_none() || nsobj_metadata.is_none() {
            return false;
        }

        let obj_metadata = obj_metadata.unwrap();
        let nsobj_metadata = nsobj_metadata.unwrap();

        // if we have modified and it is the same it must be the same file
        if obj_metadata.modified.is_some() && nsobj_metadata.modified.is_some() {
            if obj_metadata.modified.unwrap() == nsobj_metadata.modified.unwrap() {
                return true;
            }
        }

        // if size is same considere comparing hash if file is small
        if obj_metadata.size == obj_metadata.size {
            if obj_metadata.size < MAX_SIZE_TO_USE_HASH
                && nsobj_metadata.size < MAX_SIZE_TO_USE_HASH
            {
                todo!()
                // return self.get_file_hash() == other.get_file_hash();
            } else {
                return true;
            }
        } else {
            return false;
        }
    }
}
